""" Tests for the 'configs submodule. """

import inspect
import os
import re
import sys
from typing import Set
from unittest import TestCase

from fastjsonschema import validate

sys.path.append(os.path.dirname(os.path.dirname(__file__)))


def getModule(serviceName):
    """
    Get the service settings module by the service name.

    Args:
        serviceName: service name

    Returns:
        a service settings module
    """
    from configs.services import lambda_tasks_unit, lambda_unit, luna_lambda, sender, video_agent, video_manager

    serviceToModuleMap = {
        "luna-sender": sender,
        "luna-lambda": luna_lambda,
        "luna-lambda-unit": lambda_unit,
        "luna-lambda-tasks-unit": lambda_tasks_unit,
        "luna-video-manager": video_manager,
        "luna-video-agent": video_agent,
    }
    return serviceToModuleMap[serviceName]


class TestConfigsCorrectness(TestCase):
    """Test settings' correctness."""

    def testGetModule(self):
        """
        Check `SERVICE_TO_CLASS_MAP` completeness.
        """
        from configs.settings.all_settings import PYTHON_SERVICES

        for service in PYTHON_SERVICES:
            match service:
                case "luna-licenses":
                    continue
                case "luna-remote-sdk":
                    continue
                case "luna-handlers":
                    continue
                case "luna-events":
                    continue
                case "luna-image-store":
                    continue
                case "luna-python-matcher":
                    continue
                case "luna-matcher-proxy":
                    continue
                case "luna-faces":
                    continue
                case "luna-tasks":
                    continue
                case "luna-accounts":
                    continue
                case "luna-configurator":
                    continue
                case "luna-admin":
                    continue
                case "luna-api":
                    continue
                case "luna-streams-retranslator":
                    continue
                case "luna-backport3":
                    continue
                case "luna-backport4":
                    continue

            with self.subTest(service=service):
                try:
                    getModule(service)
                except KeyError as e:
                    msg = f"Add the following service to the getModule function: '{service}'"
                    raise self.failureException(msg) from e

    def testSettingsCorrectness(self):
        """
        Check each `setting` has the same name as the key in the `SETTINGS` dict.
        """
        from configs.settings.all_settings import SETTINGS

        for settingName, setting in SETTINGS.items():
            with self.subTest(setting=settingName):
                self.assertTrue(settingName.isupper(), "Set setting name upper")
                # I need private attribute, so... pylint: disable-msg=W0212
                self.assertEqual(settingName, setting._settingName, "Set setting name the same")

    def testServicesCompleteness(self):
        """
        Check `services` field in each `setting` (i.e. DATABASE_NUMBER) to be in the corresponding `service adapter`
        (i.e. `SettingsFaces`).
        """
        from configs_core.base_classes import BaseSettingClass
        from configs_core.base_service_setting import BaseServiceSettingsClass as CoreBaseServiceSettingsClass

        from configs.settings.all_settings import PYTHON_SERVICES, SETTINGS

        def getConstants(cls: type, service: str) -> Set[str]:
            """
            Get all `str` constants from the class __init__ method.

            Args:
                cls: class to scan
                service: service name

            Returns:
                constants
            """
            servicePrefix = service.upper().replace("-", "_")
            strings = {const.upper() for const in cls.__init__.__code__.co_consts if isinstance(const, str)}
            if hasattr(cls, "applySettings"):
                strings.update(
                    {const.upper() for const in cls.applySettings.__code__.co_consts if isinstance(const, str)}
                )
            if hasattr(cls, "applyEstimatorsSettings"):
                strings.update(
                    {
                        const.upper()
                        for const in cls.applyEstimatorsSettings.__code__.co_consts
                        if isinstance(const, str)
                    }
                )
            goodStrings = {string for string in strings if re.match(r"^[_a-zA-Z][a-zA-Z0-9_\-]*$", string) is not None}
            goodStrings = {
                string if not string.startswith("_") else f"{servicePrefix}{string}" for string in goodStrings
            }
            return goodStrings

        def isNeededClass(cls) -> bool:
            """
            Whether we need a class.

            Args:
                cls: class

            Returns:
                need
            """
            if not inspect.isclass(cls):
                return False
            return any(
                (
                    issubclass(cls, BaseSettingClass),
                    issubclass(cls, CoreBaseServiceSettingsClass),
                )
            )

        for settingName, setting in SETTINGS.items():
            if settingName == "LUNA_CONFIGURATOR":
                # This setting is not stored in the configurator. Guess why.
                continue
            for serviceName in set(setting.services) & set(PYTHON_SERVICES):
                match serviceName:
                    case "luna-licenses":
                        continue
                    case "luna-remote-sdk":
                        continue
                    case "luna-handlers":
                        continue
                    case "luna-events":
                        continue
                    case "luna-image-store":
                        continue
                    case "luna-python-matcher":
                        continue
                    case "luna-matcher-proxy":
                        continue
                    case "luna-faces":
                        continue
                    case "luna-tasks":
                        continue
                    case "luna-accounts":
                        continue
                    case "luna-configurator":
                        continue
                    case "luna-admin":
                        continue
                    case "luna-api":
                        continue
                    case "luna-streams-retranslator":
                        continue
                    case "luna-backport4":
                        continue
                    case "luna-backport3":
                        continue
                serviceModule = getModule(serviceName)
                with self.subTest(setting=settingName, service=serviceName):
                    constants = set()
                    for _, moduleObjectClass in inspect.getmembers(serviceModule, isNeededClass):
                        constants |= getConstants(moduleObjectClass, serviceName)
                    self.assertTrue(
                        settingName in constants,
                        f"Seems like the setting '{settingName}' " f"is not used in the '{serviceName}' service",
                    )

    def test_default_values(self):
        """
        Check default values.
        """
        from configs.settings.all_settings import SETTINGS

        for setting in SETTINGS.values():
            with self.subTest(setting_name=setting._settingName):
                validate(setting.schema, setting.default)
